{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
{
        Program Name   : Express
        Package Name   : Subroutine Library - Layer 4
        Purpose        : Expression evaluator
        Institution    : Conroy & Conroy Co.
        Date written   : 25-Aug-2002
        Written by     : Alan Conroy
        Version        : 1.0I

        Copyright(C) 2002-2016 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *******************************************************
        *                                                     *
        *      M O D I F I C A T I O N   H I S T O R Y        *
        *                                                     *
        *******************************************************

              DATE    BY      REASON

          12-Oct-2002 EAC     Fix protection violation caused
                              by trailing operator.
                              Fix handling of E-notation.
           2-Nov-2002 EAC     Handle division by 0.
                              Fix compilation problems in BP.
                              Change Get_Symbol parameters to longint.
                              Fix error in calculation of HypArcCsc.
                              Fix problems with handling E-notation.
                              Code cleanup.
                              Fix returning of error code.
          30-Nov-2002 EAC     Properly handle an expression
                              which is simply a number.
                              Fixed unary minus on functions.
                              Initialize strings for Delphi.
          4-Apr-2006  EAC     Added SHR and SHL operators.
          6-Dec-2006  EAC     Fix bug that swallowed tokens
                              when processing a function name
                              that ended with "E".
          13-Apr-2007 EAC     Fix handling of alphanumeric
                              operators when following numeric literal is not
                              separated from operator with white space.
                              Increased "D" operator precedence.
          11-Jun-2008 EAC     Added NOT right-unary operator.
          9-Apr-2012  Klaus   Fix some minor bugs, add support for string
                              functions and operators
          14-Apr-2012 EAC     Fixed bug that sometimes mangled the expression
                              stack when using functions.
          22-Sep-2016 EAC     Added SF function.

        *******************************************************
        *                                                     *
        *      P R O G R A M   P U R P O S E                  *
        *                                                     *
        *******************************************************

          This unit contains classes and routines to evaluate expressions.

          This code was culled from the Sirius compiler expression code.
}

unit Express ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF FPC_OR_DELPHI}
     SysUtils, { strdispose }
{$ELSE}
     Strings, { strdispose }
{$ENDIF}

     { Subroutine Library - Layer 0... }
     CommBase, { TBase_Object }
     CommonUt, { Low_String_Index }

     { Subroutine Library - Layer 3... }
{$IFDEF FPC_OR_DELPHI}
     UOS_Parse  { TParser }
{$ELSE}
     Parse { TSParser }
{$ENDIF}
     ;

{ Symbol flags... }
const SF_Right_Unary_Operator = 1 ;
const SF_Left_Unary_Operator = 2 ;
const SF_Binary_Operator = 3 ;
const SF_Function = 4 ;

type tNode_type = (Node_Literal, Node_Function, Node_Operator) ;
     pExpression_Node = ^tExpression_Node ;
     tExpression_Node = record
                            Node_Type : tNode_Type ; { Class of node }
                            Operator : longint ; { Operator or function value }
                            Precedence : integer ; { Precedence of node }
                            Back, Left, Right : pExpression_Node ;
                            Literal : string ;
                        end ;

type TSExpression_Evaluator = object( TBase_Object )
                                 private { Instance data... }
                                     _Error : integer ;
                                     Literal_Text : string ;
{$IFDEF FPC_OR_DELPHI}
                                     Parser : TParser ;
{$ELSE}
                                     Parser : TSParser ;
{$ENDIF}
                                     _Position : integer ; { Position of error }
                                     Stack : array[ 0..1024 ] of extended ;
                                     SStack: array[ 0..128 ] of string ;
                                     SSP : integer ;    { String Stack Pointer }
                                     Parameter_Delimiter : string ;
                                     Whole_Length : integer ;
                                     { Length of full expression to evaluate }

                                 public { Constructors and destructors... }
                                     constructor Init ;

                                 private { Internal utility routines... }
                                     function Expression : boolean ;
                                     { Generate code for an expression. }

                                     procedure Generate_Tree( X : pExpression_Node ) ;
                                     { Evaluate tree with root X. }

                                     function Get_Subexpression : pExpression_Node ;
                                     { Get a subexpression (term) and return
                                       pointer to subexpression tree }

                                     function Number : Boolean ;
                                     procedure Optimize_Tree( X : pExpression_Node ) ;
                                     function Parse_Expression : pExpression_Node ;
                                     { Parse an expression into an expression
                                       tree, and return pointer to root }

                                     procedure Zero_Expression_Tree( X : pExpression_Node ) ;

                                 public { API... }
                                     function Calculate( Operator : integer ) : boolean ;
                                         virtual ;
                                     function ERT( X : integer ) : string ;
                                         virtual ;
                                     function Evaluate( X : string ) : boolean ;
                                     function Function_Reference( Value : integer ) : boolean ;
                                         virtual ;
                                     { Process function parameters.
                                       Value = parameter count. }

                                     function Get_Position : integer ;
                                     { Position of error, if there was one. }

                                     function Get_Result : extended ;
                                     procedure Get_Symbol( X : string ;
                                         var Flags : longint ;
                                         var Address : int64 ;
                                         var Value, _Result : longint ) ;
                                         virtual ;
                                     function _Last_Error : integer ;
                                     function Pop : extended ;
                                     procedure Push( F : extended ) ;
                                     function SPop : string ;
                                     procedure SPush( ST : string ) ;
                                     function Get_SResult : string ;
                                     function Get_Dump: string ;
                                     function IsResultString : boolean ;
                                     procedure Put_Token( const X : string ) ;
                                     function Token : string ;
                                     function Grab_Line : string ;
                                     procedure Set_Parameter_Delimiter( S : string ) ;
                                     procedure Set_Decimal_Char( C : char ) ;
                                     procedure Set_Error( E : integer ) ;
                             end ; { TSExpression_Evaluator }
     PSExpression_Evaluator = ^TSExpression_Evaluator ;

function EERT( X : integer ) : string ;
procedure Eval( Expression : string ; var _Result, Imaginary : Extended ;
    var E, P : Integer ) ;


implementation

uses { Borland... }
     Math, { IntPower }

     { Subroutine Library - Layer 0... }
     L10n, { Text_* }

     { Subroutine Library - Layer 1... }
     Instrs, { Instr }
     Maths, { Power }
     Num1s, { Num1s }

     { Subroutine Library - Layer 3... }
     O_S, { OS }
     UStrings ; { PString }


var Randomized : boolean ;
    Decimal_Char : char ;


procedure Ensure_Randomness ;

begin
    if( not Randomized ) then
    begin
        randomize ;
        Randomized := True ;
    end ;
end ;


{ Operations... }
const Op_Negate_Stack = 1 ;
const Op_Add = 2 ;
const Op_Subtract = 3 ;
const Op_Multiply = 4 ;
const Op_Divide = 5 ;
const Op_Percent = 6 ;
const Op_Exponent = 7 ;
const Op_Factorial = 8 ;

const Op_Rad = 9 ;
const Op_Deg = 10 ;
const Op_Abs = 11 ;
const Op_Cos = 12 ;
const Op_Cot = 13 ;
const Op_Csc = 14 ;
const Op_Exp = 15 ;
const Op_Int = 16 ;
const Op_Log = 17 ;
const Op_Sec = 18 ;
const Op_Sin = 19 ;
const Op_Tan = 20 ;
const Op_Fix = 21 ;
const Op_Sqr = 22 ;
const Op_Sgn = 23 ;
const Op_Fib = 24 ;

const Op_Min = 25 ;
const Op_Max = 26 ;

const Op_Arccos = 27 ;
const Op_Arccot = 28 ;
const Op_Arccsc = 29 ;
const Op_Arcsec = 30 ;
const Op_Arcsin = 31 ;
const Op_Arctan = 32 ;
const Op_Hypcos = 33 ;
const Op_Hypcot = 34 ;
const Op_Hypcsc = 35 ;
const Op_Hypsec = 36 ;
const Op_Hypsin = 37 ;
const Op_Hyptan = 38 ;
const Op_Hyparccos = 39 ;
const Op_Hyparccot = 40 ;
const Op_HypArcCsc = 41 ;
const Op_Hyparcsec = 42 ;
const Op_Hyparcsin = 43 ;
const Op_Hyparctan = 44 ;
const Op_Log10 = 45 ;
const Op_Rnd = 46 ;

const Op_Mod = 47 ;
const Op_Rem = 48 ;
const Op_D = 49 ;
const Op_Div = 50 ;

const Op_Root = 51 ;
const Op_PI = 52 ;

const Op_Shift_Left = 53 ;
const Op_Shift_Right = 54 ;
const Op_And = 55 ;
const Op_Or = 56 ;
const Op_Xor = 57 ;
const Op_Complement = 58 ;
const Op_NAnd = 59 ;
const Op_NOr = 60 ;
const Op_XNor = 61 ;

const Op_Round = 62 ;

const Op_SF = 63 ;

{ Errors... }
const Error_Expected_Comma = 1 ;
const Error_Syntax_Error = 2 ;
const Error_Illegal_Symbol = 3 ;
const Error_Expected_Left_Parenthesis = 4 ;
const Error_Division_By_Zero = 5 ;
const Error_Illegal_Number_for_factorial = 6 ;
const Error_Floating_point_error = 7 ;
const Error_Illegal_Number = 8 ;
const Error_Illegal_Argument = 9 ;
const Error_Imaginary_Number = 10 ;
const Error_Illegal_Argument_in_LOG = 11 ;
const Error_Argument_Too_large_in_EXP = 12 ;
const Error_Missing_Right_Parenthesis = 13 ;
const Error_Integer_To_Large = 14 ;
const Error_Expression_Too_Complex = 15 ;
const Error_Internal_Error = 16 ;


function Numval( S : string ; var Dummy : integer ) : extended ;

var E : extended ;
    I : integer ;

begin
    // Normalize to decimal...
    if( Decimal_Char <> '.' ) then
    begin
        I := instr( 0, S, Decimal_Char ) ;
        if( I >= Low_String_Index ) then
        begin
            S[ I ] := '.' ;
        end ;
    end ;

    // Normalize to locale...
    Dummy := instr( 0, S, '.' ) ;
    if( Dummy >= Low_String_Index ) then { Found decimal point }
    begin
        S := Left_Of( S, Dummy ) +
            OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal +
            Right_Of( S, Dummy ) ;
    end ;

    // Convert...
    E := Xval( S, Dummy ) ;
    Numval := E ;
end ;


function RoundToDig(const MyValue: Extended; const NDigit: Extended = -2): Extended ;

var Decim : Extended ;
    I : integer ;
    S : string ;
    N : integer ;

begin
    N := trunc( int( NDigit ) ) ;
    if( N < 0 ) then // Round to right of decimal
    begin
        S := '' ;
        for I := N to -1 do
        begin
            S := S + '0' ;
        end ;
        S := '0.' + S + '5' ;
        Decim := MyValue + strtofloat( S ) ; // Round
        S := Num2( Decim ) ;
        I := instr( 0, S, '.' ) ;
        if( I >= Low_String_Index ) then // If we have a fractional portion
        begin
            I := I - N ; // Position of last significant digit
            S := Left( S, I + 1 - Low_String_Index ) ; // Trim result to requested number of digits
            Decim := strtofloat( S ) ;
        end ;
        Result := Decim ;
        exit ;
    end ;

    // Round to left of decimal...
    Decim := IntPower(10, N);
    if MyValue > 0 then
        Result := Trunc((MyValue / Decim) + 0.5) * Decim
    else if MyValue < 0 then
        Result := Trunc((MyValue / Decim) - 0.5) * Decim
    else
        Result := 0 ;
end ; // RoundToDig


function EERT( X : integer ) : string ;

begin
    case X of
        Error_Expected_Comma: EERT := Text_Express_Error_Expected_Comma ;
        Error_Syntax_Error: EERT := Text_Express_Error_Syntax_Error ;
        Error_Illegal_Symbol: EERT := Text_Express_Error_Illegal_Symbol ;
        Error_Expected_Left_Parenthesis: EERT := Text_Express_Error_Expected_Left_Parenthesis ;
        Error_Division_By_Zero: EERT := Text_Express_Error_Division_By_Zero ;
        Error_Illegal_Number_for_factorial: EERT := Text_Express_Error_Illegal_Number_for_factorial ;
        Error_Floating_point_error: EERT := Text_Express_Error_Floating_point_error ;
        Error_Illegal_Number: EERT := Text_Express_Error_Illegal_Number ;
        Error_Illegal_Argument: EERT := Text_Express_Error_Illegal_Argument ;
        Error_Imaginary_Number: EERT := Text_Express_Error_Imaginary_Number ;
        Error_Illegal_Argument_in_LOG: EERT := Text_Express_Error_Illegal_Argument_in_LOG ;
        Error_Argument_Too_large_in_EXP: EERT := Text_Express_Error_Argument_Too_large_in_EXP ;
        Error_Missing_Right_Parenthesis: EERT := Text_Express_Error_Missing_Right_Parenthesis ;
        Error_Integer_To_Large: EERT := 'Integer too large' ;
        Error_Expression_Too_Complex: EERT := 'Expression too complex' ;
        Error_Internal_Error: EERT := 'Internal error' ;
    end ;
end ;



{ TSExpression_Evaluator methods... }

{ Constructors and destructors... }

constructor TSExpression_Evaluator.Init ;

var
{$IFDEF DYNAMIC_STRINGS}
    I : integer ;
{$ENDIF}
    W : string ;

begin
    inherited Init ;

{$IFDEF DYNAMIC_STRINGS}
    Initialize( Literal_Text ) ;
    Initialize( Parameter_Delimiter ) ;
    for I := low( SStack ) to high( SStack ) do
    begin
        Initialize( SStack[ I ] ) ;
    end ;
{$ENDIF}
    Parameter_Delimiter := ',' ;
    Decimal_Char := '.' ;
    W := OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal ;
    if( W <> '' ) then
    begin
        Decimal_Char := W[ Low_String_Index ] ;
    end ;
end ;



{ API... }

function TSExpression_Evaluator.ERT( X : integer ) : string ;

begin
    ERT := EERT( X ) ;
end ;


procedure TSExpression_Evaluator.Get_Symbol( X : string ; var Flags : longint ;
    var Address : int64 ; var Value, _Result : longint ) ;

begin
    _Result := 0 ; { Assume success }
    X := uppercase( X ) ;
    if( X = '+' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Add ;
        Value := 16000 ; { Precedence }
        exit ;
    end ;
    if( X = '-' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Subtract ;
        Value := 16000 ; { Precedence }
        exit ;
    end ;
    if( X = '*' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Multiply ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = '/' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Divide ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = '^' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Exponent ;
        Value := 26000 ; { Precedence }
        exit ;
    end ;
    if( X = 'SHL' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Shift_Left ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'SHR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Shift_Right ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'AND' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_And ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'OR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Or ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'XOR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Xor ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'NAND' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_NAnd ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'NOR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_NOr ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'XNOR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_XNor ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'MIN' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Min ;
        Value := 10000 ; { Precedence }
        exit ;
    end ;
    if( X = 'MAX' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Max ;
        Value := 10000 ; { Precedence }
        exit ;
    end ;
    if( X = '%' ) then
    begin
        Flags := SF_Left_Unary_Operator ;
        Address := Op_Percent ;
        Value := 28000 ;
        exit ;
    end ;
    if( X = '!' ) then
    begin
        Flags := SF_Left_Unary_Operator ;
        Address := Op_Factorial ;
        Value := 28000 ;
        exit ;
    end ;
    if( X = 'NOT' ) then
    begin
        Flags := SF_Right_Unary_Operator ;
        Address := Op_Complement ;
        Value := 28000 ;
        exit ;
    end ;
    if( X = 'MOD' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Mod ; { Modulus }
        Value := 24000 ;
        exit ;
    end ;
    if( X = 'REM' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Rem ; { Remainder }
        Value := 24000 ;
        exit ;
    end ;
    if( X = 'D' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_D ;
        Value := 26000 ;
        exit ;
    end ;
    if( X = 'DIV' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Div ;
        Value := 24000 ;
        exit ;
    end ;

    if( X = 'ROOT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Root ;
        Value := 2 ;
        exit ;
    end ;
    if( X = 'RAD' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Rad ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'RND' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Rnd ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'DEG' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Deg ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ABS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Abs ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'COS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Cos ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'COT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Cot ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'CSC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Csc ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'EXP' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Exp ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'INT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Int ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'LOG' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Log ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SEC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sec ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SIN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sin ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'TAN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Tan ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'FIX' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Fix ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SQR' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sqr ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SGN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sgn ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'FIB' ) then
    begin
        Flags := SF_Function ;
        Address := Op_FIB ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCCOS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcCos ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCCOT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcCot ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCCSC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcCsc ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCSEC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcSec ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCSIN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcSin ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCTAN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcTan ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPCOS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypCos ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPCOT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypCot ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPCSC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypCsc ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPSEC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypSec ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPSIN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hypsin ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPTAN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypTan ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCCOS' ) or ( X = 'ARCHYPCOS' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparccos ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCCOT' ) or ( X = 'ARCHYPCOT' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparccot ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCCSC' ) or ( X = 'ARCHYPCSC' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_HyparcCsc ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCSEC' ) or ( X = 'ARCHYPSEC' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparcsec ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCSIN' ) or ( X = 'ARCHYPSIN' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparcsin ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCTAN' ) or ( X = 'ARCHYPTAN' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparctan ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'LOG10' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Log10 ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SF' ) then
    begin
        Flags := SF_Function ;
        Address := Op_SF ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'PI' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Pi ;
        Value := 0 ;
        exit ;
    end ;
    if( X = '#' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Round ;
        Value := 26000 ; { Precedence }
        exit ;
    end ;

    _Result := 1 ; { Error: unknown symbol }
end ; // TSExpression_Evaluator.Get_Symbol


procedure TSExpression_Evaluator.SPush( ST : string ) ;

begin
    if( SSP >= high( SStack ) ) then
    begin
        Set_Error( Error_Expression_Too_Complex ) ;
    end ;
    SSP := SSP + 1 ;
    SStack[ SSP ] := ST ;
end ;


procedure TSExpression_Evaluator.Push( F : extended ) ;

begin
    if( Stack[ 0 ] >= high( Stack ) ) then
    begin
        Set_Error( Error_Expression_Too_Complex ) ;
    end ;
    Stack[ 0 ] := Stack[ 0 ] + 1 ;
    Stack[ trunc( Stack[ 0 ] ) ] := F ;
end ;


function Cvt_String( X : PString ) : string ;

begin
    Cvt_String := X^.Text ;
    dispose( X, Done ) ;
end ;


function TSExpression_Evaluator.Function_Reference( Value : integer ) : boolean ;
{ Process function parameters.  Value = parameter count. }

var F : integer ;
    G : longint ;
    Y : string ;

begin
    Function_Reference := False ;

    if( Value <> 0 ) then { if parameters }
    begin
        Y := Token ;
        G := 0 ; { Number of parameters that code was generated for }
        if( instr( 0, '({[', Y ) < Low_String_Index ) then { No parameters specified }
        begin
            Put_Token( Y ) ; { Restore token }
            if( Value > 0 ) then
            { One or more required parameters }
            begin
                Set_Error( Error_Expected_Left_Parenthesis ) ;
                exit ;
            end ;
        end else { Parenthesis found }
        begin
            F := 1 ; { Number of parameters }
            while( F <= abs( Value ) ) do { Until out of parameters }
            begin
                { Process next parameter... }
                _Error := 0 ;
                if( not Expression ) then
                begin
                    exit ;
                end ;
                Inc( G ) ;
                Y := Token ; { Get next token }
                if(
                    ( Y <> Parameter_Delimiter )
                    and
                    ( F <= abs( Value ) )
                  ) then { No more parameters }
                begin
                    if( instr( 0, '}])', Y ) < Low_String_Index ) then
                    begin
                        Put_Token( Y ) ;
                    end ;
                    if( Value < 0 ) then { Variable number of parameters }
                    begin
                        F := abs( Value ) + 1 ; { Get actual number }
                    end else
                    if( F < Value ) then { There should be more parameters }
                    begin
                        Set_Error( Error_Expected_Comma ) ;
                        exit ;
                    end ;
                end ;
                Inc( F ) ; { Increment parameter count }
            end ; {while( F <= abs( Value ) ) do}
        end ; {if}
        if( Value < 0 ) then { Variable number of parameters - push count of them }
        begin
            Push( G ) ;
        end ;
        if( ( instr( 0, ']})', Y ) < Low_String_Index ) and ( G > 0 ) ) then { No parenthesis when needed }
        begin
            Set_Error( Error_Missing_right_parenthesis ) ;
            exit ;
        end ;
    end ; { if Last_Symbol^.Size <> 0 }
    Function_Reference := True ;
end ; { Function_Reference }


procedure TSExpression_Evaluator.Zero_Expression_Tree( X : pExpression_Node ) ;
{ Zero storage stored at tree with root X }

var Current, Previous : pExpression_Node ;

begin
    { Start at root }
    Current := X ;
    while( Current <> nil ) do
    begin
        { Go all the way left }
        while( Current^.Left <> nil ) do
        begin
            Current := Current^.Left ;
        end ;
        if( Current^.Right <> nil ) then
        begin
            Current := Current^.Right ;
        end else
        begin
            Previous := Current^.Back ;
            if( Previous <> nil ) then
            begin
                if( Previous^.Left = Current ) then
                begin
                    Previous^.Left := nil ;
                end else
                begin
                    Previous^.Right := nil ;
                end ;
            end ;
            case Current^.Node_Type of
{$IFDEF FPC_OR_DELPHI}
                Node_Literal : finalize( Current^.Literal ) ;
{$ELSE}
                Node_Literal : strdispose( Current^.Literal ) ;
{$ENDIF}
            end ;
            dispose( Current ) ;
            Current := Previous ;
        end ;
    end ; { while( Current <> nil ) do }
end ; { Zero_Expression_Tree }


procedure TSExpression_Evaluator.Set_Error( E : integer ) ;

begin
    if( _Error = 0 ) then { Not already set }
    begin
        _Error := E ;
    end ;
end ;


procedure TSExpression_Evaluator.Optimize_Tree( X : pExpression_Node ) ;
{ Optimize an expression tree. }

var Dummy : integer ;
    L, R : string ;
    LX, RX, Temp : extended ;
{$IFDEF INT64}
    LI, RI : int64 ;
{$ELSE}
    LI, RI : longint ;
{$ENDIF}
    Res : integer ;

begin
    { Optimize sub-trees }
    if( X^.Left <> nil ) then
    begin
        Optimize_Tree( X^.Left ) ;
    end ;
    if( X^.Right <> nil ) then
    begin
        Optimize_Tree( X^.Right ) ;
    end ;

    { See if we can optimize... }
    if( X^.Node_Type <> Node_Operator ) then
    begin
        exit ; { Can't optimize this any further }
    end ;
    LX := 0 ; { Avoid compiler warning }
    RX := 0 ; { Ditto }
    if( X^.Left <> nil ) then
    begin
        if( X^.Left^.Node_Type <> Node_Literal ) then
        begin
            exit ;
        end ;
        L := X^.Left^.Literal ;
        LX := Numval( L, Dummy ) ;

    end ;
    if( X^.Right <> nil ) then
    begin
        if( X^.Right^.Node_Type <> Node_Literal ) then
        begin
            exit ;
        end ;
        R := X^.Right^.Literal ;
        RX := Numval( R, Dummy ) ;
    end ;

    case X^.Operator of
        Op_Negate_Stack : LX := -LX ;
        Op_Add : LX := LX + RX ;
        Op_Subtract : LX := LX - RX ;
        Op_Multiply : LX := LX * RX ;
        Op_Divide :
            begin
                if( RX = 0.0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                end else
                begin
                    LX := LX / RX ;
                end ;
            end ;
        Op_Shift_Right :
            begin
                if( RX >= 1 ) then
                begin
                    Temp := 2 ;
                    while( RX >= 2 ) do
                    begin
                        RX := RX - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    LX := LX / Temp ;
                end ;
            end ;
        Op_Shift_Left :
            begin
                if( RX >= 1 ) then
                begin
                    Temp := 2 ;
                    while( RX >= 2 ) do
                    begin
                        RX := RX - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    LX := LX * Temp ;
                end ;
            end ;
        Op_And :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := LI and RI ;
                    LX := LI ;
                end ;
            end ;
        Op_Or :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := LI or RI ;
                    LX := LI ;
                end ;
            end ;
        Op_Xor :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := LI xor RI ;
                    LX := LI ;
                end ;
            end ;
        Op_NAnd :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := not( LI and RI ) ;
                    LX := LI ;
                end ;
            end ;
        Op_NOr :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := not( LI or RI ) ;
                    LX := LI ;
                end ;
            end ;
        Op_XNor :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := not( LI xor RI ) ;
                    LX := LI ;
                end ;
            end ;
        Op_Percent : LX := LX / 100.0 ;
        Op_Exponent :
            begin
                LX := Power( LX, RX, Res ) ;
                Set_Error( Res ) ;
            end ;
        Op_Factorial : LX := Factorial( LX ) ;
        Op_Complement :
            begin
                if(
                    ( int( RX ) > high( LI ) )
                    or
                    ( int( RX ) < low( LI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                end else
                begin
                    LI := trunc( int( RX ) ) ;
                    LI := not LI ;
                    LX := LI ;
                end ;
            end ;
        Op_Max :
            begin
                if( LX < RX ) then LX := RX ;
            end ;
        Op_Min :
            begin
                if( LX > RX ) then LX := RX ;
            end ;
        else exit ;
    end ; { case X.Operator of }

    L := num2( LX ) ;
    X^.Literal := L ;
    X^.Node_Type := Node_Literal ;
    if( X^.Left <> nil ) then
    begin
{$IFDEF FPC_OR_DELPHI}
        finalize( X^.Left^.Literal ) ;
{$ENDIF}
        dispose( X^.Left ) ;
        X^.Left := nil ;
    end ;
    if( X^.Right <> nil ) then
    begin
{$IFDEF FPC_OR_DELPHI}
        finalize( X^.Right^.Literal ) ;
{$ENDIF}
        dispose( X^.Right ) ;
        X^.Right := nil ;
    end ;
end ; { Optimize_Tree }


function TSExpression_Evaluator.Pop : extended ;

begin
    Pop := Stack[ trunc( Stack[ 0 ] ) ] ;
    if( Stack[ 0 ] = 0 ) then
    begin
        Set_Error( Error_Internal_Error ) ;
    end else
    begin
        Stack[ 0 ] := Stack[ 0 ] - 1 ;
    end ;
end ;


function TSExpression_Evaluator.SPop : string ;

begin
    SPop := SStack[ SSP ] ;
    if( SSP = 0 ) then
    begin
        Set_Error( Error_Internal_Error ) ;
    end else
    begin
        SSP := SSP - 1 ;
    end ;
end ;


function TSExpression_Evaluator.Calculate( Operator : integer ) : boolean ;

var Dummy : integer ;
    A1, A2, A3, E1, E2, Temp : extended ;
{$IFDEF INT64}
    LI, RI : int64 ;
{$ELSE}
    LI, RI : longint ;
{$ENDIF}
    Res : integer ;

begin
    Calculate := False ; { Assume no errors }
    case Operator of
        Op_Negate_Stack :
            begin
                Push( -Pop ) ;
            end ;
        Op_Add :
            begin
                Push( Pop + Pop ) ;
            end ;
        Op_Subtract :
            begin
                E2 := Pop ;
                E1 := Pop ;
                Push( E1 - E2 ) ;
            end ;
        Op_Multiply :
            begin
                Push( Pop * Pop ) ;
            end ;
        Op_Divide :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0.0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( E1 / E2 ) ;
            end ;
        Op_Shift_Left :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 >= 1 ) then
                begin
                    Temp := 2 ;
                    while( E2 >= 2 ) do
                    begin
                        E2 := E2 - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    Push( E1 * Temp ) ;
                end ;
            end ;
        Op_Shift_Right :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 >= 1 ) then
                begin
                    Temp := 2 ;
                    while( E2 >= 2 ) do
                    begin
                        E2 := E2 - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    Push( E1 / Temp ) ;
                end ;
            end ;
        Op_And :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                    or
                    ( int( E2 ) > high( RI ) )
                    or
                    ( int( E2 ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                RI := trunc( int( E2 ) ) ;
                LI := LI and RI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Or :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                    or
                    ( int( E2 ) > high( RI ) )
                    or
                    ( int( E2 ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                RI := trunc( int( E2 ) ) ;
                LI := LI or RI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Xor :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                    or
                    ( int( E2 ) > high( RI ) )
                    or
                    ( int( E2 ) < low( RI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                RI := trunc( int( E2 ) ) ;
                LI := LI xor RI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Percent :
            begin
                Push( Pop / 100.0 ) ;
            end ;
        Op_Exponent :
            begin
                E2 := Pop ;
                E1 := Pop ;
                E1 := Power( E1, E2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( E1 ) ;
            end ;
        Op_Round :
            begin
                E2 := Pop ;
                E1 := Pop ;
                E1 := RoundToDig( E1, E2 ) ;
                Push( E1 ) ;
            end ;
        Op_Factorial :
            begin
                Push( Factorial( Pop ) ) ;
            end ;
        Op_Complement :
            begin
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                  ) then
                begin
                    Set_Error( Error_Integer_To_Large ) ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                LI := not LI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Max :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E1 > E2 ) then
                begin
                    E1 := E2 ;
                end ;
                Push( E1 ) ;
            end ;
        Op_Min :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E1 < E2 ) then
                begin
                    E1 := E2 ;
                end ;
                Push( E1 ) ;
            end ;

        Op_Rad :
            begin
                Push( Pop / ( 180. / Pi ) ) ;
            end ;
        Op_Deg :
            begin
                Push( Pop * ( 180. / Pi ) ) ;
            end ;
        Op_Abs :
            begin
                Push( Abs( Pop ) ) ;
            end ;
        Op_Cos :
            begin
                Push( Cos( Pop ) ) ;
            end ;
        Op_Cot :
            begin
                Push( 1 / Tan( Pop ) ) ;
            end ;
        Op_Csc :
            begin
                Push( 1. / Sin( Pop ) ) ;
            end ;
        Op_Exp :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Exp( E1 ) ) ;
            end ;
        Op_Int :
            begin
                Push( Int( Pop ) ) ;
            end ;
        Op_Log :
            begin
                E1 := Pop ;
                if( E1 <= 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Ln( E1 ) ) ;
            end ;
        Op_Sec :
            begin
                Push( 1. / Cos( Pop ) ) ;
            end ;
        Op_Sin :
            begin
                Push( Sin( Pop ) ) ;
            end ;
        Op_Tan :
            begin
                Push( Tan( Pop ) ) ;
            end ;
        Op_Fix :
            begin
                Push( Trunc( Pop ) ) ;
            end ;
        Op_Sqr :
            begin
                E1 := Pop ;
                if( E1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Sqrt( E1 ) ) ;
            end ;
        Op_Sgn :
            begin
                Push( Sgn( Pop ) ) ;
            end ;
        Op_Fib :
            begin
                E1 := Pop ;
                if( E1 > 182 ) then { No fibonacci > 182 }
                begin
                    Set_Error( Error_Floating_point_error ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Int( Abs( E1 ) ) ;
                if( E1 < 2 ) then
                begin
                    Push( 1 ) ;
                end else
                begin
                    A1 := 1 ;
                    A2 := 1 ;
                    for Dummy := 2 to Trunc( E1 ) do
                    begin
                        A3 := A2 ;
                        A2 := A1 ;
                        A1 := A2 + A3
                    end ;
                    Push( A1 + A2 ) ;
                end
            end ;
        Op_Arccos :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := 1 - A1 ;
                if( A1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( Sqrt( A1 ) / E1 ) ) ;
            end ;
        Op_ArcCot :
            begin
                E1 := Pop ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( 1. / E1 ) ) ;
            end ;
        Op_ArcCsc :
            begin
                E1 := Power( Pop, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 - 1 ;
                if( E1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( 1. / Sqrt( E1 ) ) ) ;
            end ;
        Op_ArcSec :
            begin
                E1 := Power( Pop, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 - 1 ;
                if( E1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( Sqrt( E1 ) ) ) ;
            end ;
        Op_Arcsin :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := 1 - A1 ;
                if( A1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( A1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    exit ;
                end ;
                Push( Arcsin( E1 ) ) ;
            end ;
        Op_ArcTan :
            begin
                Push( Arctan( Pop ) ) ;
            end ;
        Op_Hypcos :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Hypcos( E1 ) ) ;
            end ;
        Op_HypCot :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := Exp( E1 ) - Exp( -E1 ) ;
                if( A1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypCot( E1 ) ) ;
            end ;
        Op_Hypcsc :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Exp( E1 ) - Exp( -E1 ) ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( 2. / E1 ) ;
            end ;
        Op_HypSec :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Exp( E1 ) + Exp( -E1 ) ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( 2. / E1 ) ;
            end ;
        Op_HypSin :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypSin( E1 ) ) ;
            end ;
        Op_HypTan :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    Set_Error( Error_Argument_Too_large_in_EXP ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := Exp( E1 ) + Exp( -E1 ) ;
                if( A1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypTan( E1 ) ) ;
            end ;
        Op_HypArcCos :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := A1 - 1 ;
                if( A1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 + Sqrt( A1 ) ;
                if( E1 <= 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Ln( E1 ) ) ;
            end ;
        Op_SF : // Superfactorial
            begin
                E1 := Pop ;
                Push( SuperFactorial( E1 ) ) ;
            end ;
        Op_HypArcCot :
            begin
                E1 := Pop ;
                if( E1 - 1 <= 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcCot( E1 ) ) ;
            end ;
        Op_HypArcCsc :
            begin
                E1 := Pop ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 < 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcCsc( E1 ) ) ;
            end ;
        Op_HypArcSec :
            begin
                E1 := Pop ;
                if( E1 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 < 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := 1 / A1 - 1 ;
                if( A1 < 0 ) then
                begin
                    Set_Error( 10 ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcSec( E1 ) ) ;
            end ;
        Op_HypArcSin :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) + 1 ;
                if( Res <> 0 ) then
                begin
                    Set_Error( Res ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( A1 < 0 ) then
                begin
                    Set_Error( Error_Imaginary_Number ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 + Sqrt( A1 ) ;
                if( E1 <= 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Ln( E1 ) ) ;
            end ;
        Op_HypArcTan:
            begin
                E1 := Pop ;
                if( abs( E1 ) >= 1 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcTan( E1 ) ) ;
            end ;
        Op_Log10 :
            begin
                E1 := Pop ;
                if( E1 <= 0 ) then
                begin
                    Set_Error( Error_Illegal_Argument_in_LOG ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Log10( E1 ) ) ;
            end ;
        Op_PI :
            begin
                Push( Pi ) ;
            end ;
        Op_Rnd :
            begin
                Ensure_Randomness ;
                Push( random( trunc( Pop ) ) ) ;
            end ;
        Op_Root :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 <> int( E2 ) ) then
                begin
                    Set_Error( Error_Illegal_argument ) ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 < 0 ) then { Negative value }
                begin
                    if( ( trunc( E2 ) and 1 ) = 0 ) then { Even number }
                    begin
                        Set_Error( Error_Imaginary_number ) ;
                        Calculate := True ;
                        exit ;
                    end ;
                end ;
                Push( Root( E1, trunc( E2 ) ) ) ;
            end ;

        Op_Mod :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( E1 - Int( E1 / E2 ) * E2 ) ;
            end ;
        Op_Rem :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 / E2 ;
                Push( E1 - Trunc( E1 ) ) ;
            end ;
        Op_Div :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0 ) then
                begin
                    Set_Error( Error_Division_By_Zero ) ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Int( Int( E1 ) / Int( E2 ) ) ) ;
            end ;
        Op_D :
            begin
                Ensure_Randomness ;
                E2 := Pop ;
                E1 := Pop ;
                A2 := 0 ;
                while( E1 > 0 ) do
                begin
                    A2 := A2 + random( trunc( E2 ) ) + 1 ;
                    E1 := E1 - 1 ;
                end ;
                Push( A2 ) ;
            end ;
    end ;
end ; { TSExpression_Evaluator.Calculate }


procedure TSExpression_Evaluator.Generate_Tree( X : pExpression_Node ) ; { Evaluate tree with root X. }

var Current, Previous : pExpression_Node ;
    Dummy : integer ;
    F : extended ;
    S : string ;

begin
    { Start at root }
    Current := X ;
    while( Current <> nil ) do
    begin
        { Go all the way left }
        while( Current^.Left <> nil ) do
        begin
            Current := Current^.Left ;
        end ;

        { Go right if no more left }
        if( Current^.Right <> nil ) then
        begin
            Current := Current^.Right ;
        end else
        begin
            { Process terminal node }
            Previous := Current^.Back ;
            if( Previous <> nil ) then
            begin
                if( Previous^.Left = Current ) then { On left of parent }
                begin
                    Previous^.Left := nil ;
                end else
                begin
                    Previous^.Right := nil ;
                end ;
            end ;
            case Current^.Node_Type of
                Node_Operator, Node_Function :
                    begin
                        if( Calculate( Current^.Operator ) ) then
                        begin
                            Stack[ 0 ] := 0 ;
                            exit ;
                        end ;
                    end ;
                Node_Literal : begin
                                   S := Current^.Literal ;
                                   F := Numval( S, Dummy ) ;
                                   Push( F ) ;
{$IFDEF FPC_OR_DELPHI}
                                   finalize( Current^.Literal ) ;
{$ENDIF}
                               end ;
            end ;
            dispose( Current ) ;
            Current := Previous ;
        end ;
    end ; { while( Current <> nil ) do }
end ; { Generate_Tree }


function TSExpression_Evaluator.Number : Boolean ;

var
{$IFDEF FUNCTION_RESULT_REQUIRED}
    A : extended ;
{$ENDIF}
    B : Integer ;
    Negate : boolean ;
    S, Saved : string ;

Label Loop ;

begin
{$IFDEF FPC_OR_DELPHI}
    Saved := Parser.Non_Terminals ;
{$ELSE}
    Saved := Parser.Non_Terminals.Text ;
{$ENDIF}
    {Parser.Non_Terminals := '0123456789.E' ;}
    Literal_Text := Token ;
    Negate := False ;
    while( Literal_Text = '+' ) do { Eat unary plus }
    begin
        Literal_Text := Token ;
    end ;
    while( Literal_Text = '-' ) do { Accumulate unary minus }
    begin
        Negate := not Negate ;
        Literal_Text := Token ;
    end ;
    if( uppercase( Literal_Text ) = 'EXP' ) then
    begin
        Number := False ;
        Put_Token( Literal_Text ) ;
        exit ;
    end ;
    if( Literal_Text = Decimal_Char ) then
    begin
        Literal_Text := '0' + Decimal_Char ;
    end ;
    if( RightMost( Literal_Text, 1 ) = 'E' ) then
    begin
        S := Token ;
        if( ( S = '-' ) or ( S = '+' ) ) then
        begin
            Literal_Text := Literal_Text + S + Token ;
        end else
        begin
            Put_Token( S ) ;
        end ;
    end ;
Loop:
{$IFDEF FUNCTION_RESULT_REQUIRED}
  {$HINTS OFF}
  {$WARNINGS OFF}
    A :=
{$ENDIF}
      Numval( Literal_Text, B ) ; { See if a valid numer }
{$IFDEF FUNCTION_RESULT_REQUIRED}
  {$WARNINGS ON}
  {$HINTS ON}
{$ENDIF}
    if( B < Low_String_Index ) then { So far, so good }
    begin
        if( not Parser.Token_EOL ) then { End of line is end of token }
        begin
            B := High_String_Index( Literal_Text ) ;
            Literal_Text := Literal_Text + Token ; { See if some more of a numeric specification }
            if( RightMost( Literal_Text, 1 ) = 'E' ) then
            begin
                S := Token ;
                if( ( S = '-' ) or ( S = '+' ) ) then
                begin
                    Literal_Text := Literal_Text + S + Token ;
                end ;
            end ;
            if( B <> High_String_Index( Literal_Text ) ) then { Source file didn't end }
            begin
                goto Loop ;
            end ;
        end ;
    end ;
    if( Left( Literal_Text, 1 ) = 'E' ) then
    begin
        B := 1 + Low_String_Index ;
    end ;
    if( ( B > Low_String_Index ) and ( Literal_Text[ B - 1 ] = 'E' ) ) then
    begin
        dec( B ) ;
    end ;
    if( B >= Low_String_Index ) then
    begin
        Put_Token( Right( Literal_Text, B ) ) ;
        Literal_Text := Left_Of( Literal_Text, B ) ;
    end ;
    if( B = Low_String_Index ) then // First character is invalid
    begin
        Number := False ;
    end else
    begin
        if( Negate ) then
        begin
            Literal_Text := '-' + Literal_Text ;
        end ;
        Number := True ;
    end ;
{$IFDEF FPC_OR_DELPHI}
    Parser.Non_Terminals := Saved ;
{$ELSE}
    Parser.Non_Terminals.Assign( Temp_String( Saved ) ) ;
{$ENDIF}
end ; { TSExpression_Evaluator.Number }


function TSExpression_Evaluator.Get_Subexpression : pExpression_Node ;
{ Get a subexpression (term) and return pointer to subexpression tree }

label Loop, Operator_Loop ;

var Subexpression, Temp : pExpression_Node ;
    Operator : pExpression_Node ;
    X : string ;
    AA : array[ 0..255 ] of char ;
    Address : int64 ;
    Flags : longint ;
    _Result, Value : longint ;

begin
    Subexpression := nil ;
Loop:
    X := Token ; { Get next token }
    if( length( X ) = 0 ) then
    begin
        Set_Error( Error_Syntax_Error ) ;
        Get_Subexpression := Subexpression ;
        exit ;
    end ;
    if( instr( 0, '({[', Left( X, 1 ) ) >= Low_String_Index ) then { Start of parenthesis}
    begin
        Temp := Parse_Expression ;
        if( Subexpression = nil ) then { First part of subexpression }
        begin
            Subexpression := Temp ;
        end else
        begin
            Operator := Subexpression ;
            while( Operator^.Left <> nil ) do
            begin
                Operator := Operator^.Left ;
            end ;
            Operator^.Left := Subexpression ;
        end ;
        X := Token ;
        if( ( X = '' ) or ( instr( 0, ')}]', Left( X, 1 ) ) < Low_String_Index ) ) then { No closing parenthesis}
        begin
            Put_Token( X ) ;
            Set_Error( Error_Missing_Right_Parenthesis ) ;
            Get_Subexpression := nil ;
            Zero_Expression_Tree( Subexpression ) ;
            exit ;
        end ;
        goto Operator_Loop ;
    end ;

    if( X = '+' ) then { Unary plus }
    begin
        goto Loop ; { Ignore these }
    end ;

    if( X = '-' ) then { Unary minus }
    begin
        Temp := Get_Subexpression ; { Get subexpression }
        if( Temp = nil ) then
        begin
            Set_Error( Error_Syntax_Error ) ;
            Get_Subexpression := Subexpression ;
            exit ;
        end ;
        new( Operator ) ;
{$IFDEF FPC_OR_DELPHI}
        initialize( Operator.Literal ) ;
{$ENDIF}
        Operator^.Back := nil ;
        Operator^.Left := Temp ;
        Operator^.Right := nil ;
        Operator^.Node_Type := Node_Operator ;
        Operator^.Precedence := 28000 ;
        Operator^.Operator := Op_Negate_Stack ;
        Temp^.Back := Operator ;
        if( Subexpression = nil ) then { Nothing in tree yet }
        begin
            Subexpression := Operator ;
        end else
        begin
            Temp := Subexpression ;
            while( Temp^.Left <> nil ) do
            begin
                Temp := Temp^.Left ;
            end ;
            Temp^.Left := Operator ;
            Operator^.Back := Temp ;
        end ;
        goto Operator_Loop ;
    end ;

    Put_Token( X ) ;
    if( Number ) then { Numeric literal }
    begin
        new( Temp ) ;
{$IFDEF FPC_OR_DELPHI}
        initialize( Temp.Literal ) ;
{$ENDIF}
        Temp^.Node_Type := Node_Literal ;
        X := Literal_Text ;
        Temp^.Left := nil ;
        Temp^.Right := nil ;
        Temp^.Back := nil ;
        Temp^.Literal := Literal_Text ;
        Subexpression := Temp ;
        goto Operator_Loop ;
    end ;

    { Must be a variable, function, or an error }
    X := Token ;
    Get_Symbol( X, Flags, Address, Value, _Result ) ;
    if( _Result <> 0 ) then
    begin
        Set_Error( Error_Syntax_Error ) ;
        Get_Subexpression := Subexpression ;
        exit ;
    end ;
    if( ( Flags and SF_Binary_Operator ) = SF_Right_Unary_Operator ) then
    begin
        new( Operator ) ;
{$IFDEF FPC_OR_DELPHI}
        initialize( Operator.Literal ) ;
{$ENDIF}
        Operator^.Back := nil ;
        Operator^.Left := nil ;
        Operator^.Node_Type := Node_Operator ;
        Operator^.Precedence := Value ;
        Operator^.Operator := Address ;
        Temp := Get_Subexpression ; { Get subexpression }
        if( Temp = nil ) then
        begin
            Set_Error( Error_Syntax_Error ) ;
            Get_Subexpression := Subexpression ;
            exit ;
        end ;
        Operator^.Right := Temp ;
        Temp^.Back := Operator ;
        if( Subexpression = nil ) then { Nothing in tree yet }
        begin
            Subexpression := Operator ;
        end else
        begin
            Temp := Subexpression ;
            while( Temp^.Left <> nil ) do
            begin
                Temp := Temp^.Left ;
            end ;
            Temp^.Left := Operator ;
            Operator^.Back := Temp ;
        end ;
        goto Operator_Loop ;
    end ;
    if( ( Flags and SF_Function ) <> 0 ) then
    { Function reference.  Function references have two pieces of code:
      1) invocation of the function (including the call, and creation of
         temporary string results), and 2) parameter passing. }
    begin
        { Evaluate function... }
        Function_Reference( Value ) ; { Process function parameters }
        Calculate( Address ) ; { Now do function evaluation }

        { Create new expression node for function call... }
        new( Temp ) ;
{$IFDEF FPC_OR_DELPHI}
        initialize( Temp.Literal ) ;
{$ENDIF}
        Temp^.Node_Type := Node_Literal ;
        X := Num2( Pop ) ; { Pop function value from stack }
        Temp^.Left := nil ;
        Temp^.Right := nil ;
        Temp^.Back := nil ;
        Temp^.Literal := X ;
        Subexpression := Temp ;
        goto Operator_Loop ;
    end ;
    Put_Token( X ) ;
    X := Token ; { Pull back from token stream }

    new( Temp ) ;
{$IFDEF FPC_OR_DELPHI}
    initialize( Temp.Literal ) ;
{$ENDIF}
    Temp^.Node_Type := Node_Function ;
    Temp^.Back := nil ;
    Temp^.Left := nil ;
    Temp^.Right := nil ;
    Temp^.Operator := Address ;
    if( Subexpression = nil ) then { Nothing in tree yet }
    begin
        Subexpression := Temp ;
    end else
    begin
        Operator := Subexpression ;
        while( Operator^.Left <> nil ) do
        begin
            Operator := Operator^.Left ;
        end ;
        Operator^.Left := Temp ;
        Temp^.Back := Operator ;
    end ;

    { Process all left-unary operators }
Operator_Loop:
    X := Token ;
    if( length( X ) > 0 ) then
    begin
        Get_Symbol( X, Flags, Address, Value, _Result ) ;
        if(
            ( _Result = 0 )
            and
            ( ( Flags and SF_Binary_Operator ) = SF_Left_Unary_Operator )
          ) then { Left-unary operator }
        begin
            { Append unary operator to subexpression tree }
            new( Operator ) ;
{$IFDEF FPC_OR_DELPHI}
            initialize( Operator.Literal ) ;
{$ENDIF}
            Operator^.Precedence := Value ;
            Operator^.Left := Subexpression ;
            Operator^.Right := nil ;
            Operator^.Back := nil ;
            Subexpression^.Back := Operator ;
            Operator^.Node_Type := Node_Operator ;
            Operator^.Operator := Address ;
            Subexpression := Operator ;
            goto Operator_Loop ; { Go back for next unary operator }
        end ;
        Put_Token( X ) ; { Push token }
    end ;

    Get_Subexpression := Subexpression ;
end ; { TSExpression_Evaluator.Get_Subexpression }


function TSExpression_Evaluator.Parse_Expression : pExpression_Node ;
{ Parse an expression into an expression tree, and return pointer to root }

label Operator_Loop ;

var Operator_Precedence : integer ;
    Operator : string ;
    Left_Subexpression, Current, Previous, Root : pExpression_Node ;
    Flags, Value : longint ;
    Address : int64 ;
    _Result : longint ;
    S : string ;

begin
    { Perform expression setup }
    Current := nil ;
    Root := nil ;

    { Get first subexpression }
    Left_Subexpression := Get_Subexpression ;

    { Process until end of expression }
    while( Left_Subexpression <> nil ) do
    begin
Operator_Loop:
        { Get operator }
        Operator := Token ;
        if( length( Operator ) = 0 ) then
        begin
            _Result := 1 ;
            Flags := 0 ;
        end else
        begin
            Get_Symbol( Operator, Flags, Address, Value, _Result ) ;
        end ;

        { This code handles the case where an operator is alphanumeric and
          not separated from following numeric literal by a space. }
        S := Operator ;
        while(
               ( length( S ) > 0 )
               and
               ( instr( 0, '0123456789', RightMost( S, 1 ) ) >= Low_String_Index )
               and
               ( _Result <> 0 )
             ) do
        begin
            S := Left( S, length( S ) - 1 ) ; { Trim last digit }
            Get_Symbol( S, Flags, Address, Value, _Result ) ;
            if( _Result = 0 ) then
            begin
                if( ( Flags and SF_Binary_Operator ) = SF_Binary_Operator ) then
                begin
                    Put_Token( Right( Operator, length( S ) + Low_String_Index ) ) ;
                    { Return digits }
                    Operator := Left( Operator, length( S ) ) ; { Strip digits }
                end ;
            end ;
        end ; // while

        if(
            ( _Result <> 0 )
            or
            ( ( Flags and SF_Binary_Operator ) = 0 )
          ) then { Invalid operator }
        begin
            Put_Token( Operator ) ; { Return token }
            if( Current <> nil ) then
            begin
                Current^.Right := Left_Subexpression ;
                Left_Subexpression^.Back := Current ;
            end else
            begin
                Root := Left_Subexpression ;
            end ;
            Set_Error( Error_Syntax_Error ) ;
            Parse_Expression := Root ;
            exit ;
        end ;
        if( ( Flags and SF_Binary_Operator ) <> SF_Binary_Operator ) then
        { Unary operator or missing binary operator }
        begin
            if( ( Flags and SF_Binary_Operator ) <> SF_Left_Unary_Operator ) then
            { Left-unary is OK }
            begin
                Put_Token( Operator ) ; { Return token }
                if( Current = nil ) then
                begin
                    Parse_Expression := Left_Subexpression ;
                end else
                begin
                    Current^.Right := Left_Subexpression ;
                    Left_Subexpression^.Back := Current ;
                    Parse_Expression := Root ;
                end ;
                Set_Error( Error_Syntax_Error ) ;
                exit ;
            end ;
        end ;

        { Create node for operator }
        if( Root = nil ) then { First thing in tree }
        begin
            new( Root ) ;
{$IFDEF FPC_OR_DELPHI}
            initialize( Root.Literal ) ;
{$ENDIF}
            Current := Root ;
            Previous := nil ;
        end else
        begin
            Previous := Current ; { Save previous current node }
            new( Current ) ;
{$IFDEF FPC_OR_DELPHI}
            initialize( Current.Literal ) ;
{$ENDIF}
        end ;
        Current^.Precedence := Value ;
        Current^.Left := Left_Subexpression ;
        Current^.Right := nil ;
        Current^.Back := Previous ;
        Current^.Node_Type := Node_Operator ;
        Current^.Operator := Address ;
        Left_Subexpression^.Back := Current ;

        { Insert operator node into existing tree }
        if( Previous <> nil ) then { Tree exists }
        begin
            if( Previous^.Left = nil ) then
            begin
                Previous^.Left := Current ;
            end else
            begin
                Previous^.Right := Current ;
            end ;
        end ; { if( Previous <> nil ) then }

        { Handle precedence }
        if( _Result = 0 ) then
        begin
            Operator_Precedence := Value ;
            if( ( Current^.Back <> nil ) and { There was a previous operator }
                ( Current^.Back^.Precedence >= Operator_Precedence )
                { Current operator has less precedence than the last
                  (Operators with same precedence evaluate from left to right,
                   thus, the following has an implied lower precedence) }
               ) then
            begin
                { Back up to the proper node }
                Previous := Current ;
                while(
                       ( Previous^.Back <> nil )
                       and
                       ( Previous^.Back^.Precedence >= Operator_Precedence )
                     ) do
                begin
                    Previous := Previous^.Back ;
                end ;

                { Move current node to proper location in tree... }
                { Unlink from current location }
                if( Current^.Back^.Left = Current ) then
                begin
                    Current^.Back^.Left := Current^.Left ;
                end else
                begin
                    Current^.Back^.Right := Current^.Left ;
                end ;
                if( Current^.Left <> nil ) then
                begin
                    Current^.Left^.Back := Current^.Back ;
                end ;
                Current^.Back := Previous^.Back ;

                { Relink to new location }
                if( Previous = Root ) then { Current becomes root node }
                begin
                    Current^.Left := Root ;
                    Root^.Back := Current ;
                    Root := Current ;
                    Current^.Back := nil ;
                end else
                begin
                    if( Previous^.Back^.Left = Previous ) then
                    begin
                        Previous^.Back^.Left := Current ;
                    end else
                    begin
                        Previous^.Back^.Right := Current ;
                    end ;
                    Current^.Left := Previous ;
                    Previous^.Back := Current ;
                end ;
            end ; { Current operator has less precedence than the last }
        end ; { if( _Result = 0 ) }

        { Get next subexpression and loop }
        Left_Subexpression := Get_Subexpression ;
    end ; { while( Left_Subexpression <> nil ) do }
    Parse_Expression := Root ;
end ; { TSExpression_Evaluator.Parse_Expression }


function TSExpression_Evaluator.Expression : boolean ; { Generate code for an expression. }

var _Result : pExpression_Node ;

begin
    _Result := Parse_Expression ;
    Expression := ( _Result <> nil  ) ;
    if( _Result <> nil ) then
    begin
        { Optimize the expression tree }
        _Error := 0 ;
        Optimize_Tree( _Result ) ;
        if( _Error <> 0 ) then { Some error during optimization }
        begin
            exit ;
        end ;

        { Generate code }
        Generate_Tree( _Result ) ;
    end ;
end ; { TSExpression_Evaluator.Expression }


function TSExpression_Evaluator.Evaluate( X : string ) : boolean ;

var Dummy : integer ;
    E : extended ;

begin
    { Setup... }
    Stack[ 0 ] := 0 ; { Clear stack }
    SSP := 0 ;
    _Error := 0 ;

    E := Numval( X, Dummy ) ;
    if( Dummy < Low_String_Index ) then { A numeric constant }
    begin
        Push( E ) ;
        Evaluate := True ;
        exit ;
    end ;

{$IFDEF FPC_OR_DELPHI}
    Parser := TParser.Create ;
{$ELSE}
    Parser.Init ;
{$ENDIF}
    Put_Token( X ) ;
    Whole_Length := length( X ) ;
    Evaluate := Expression ;
    if(
        ( _Error = 0 ) { Haven't already flagged an error }
        and
        (
          ( Token <> '' )
          or
          ( ( Stack[ 0 ] <> 1 ) and ( SSP = 0 ) )
        )
      ) then { Left-over unparsable text }
    begin
        Set_Error( Error_Syntax_Error ) ;
    end ;
    if( _Error <> 0 ) then
    begin
        Evaluate := False ;
    end ;
    _Position := Whole_Length - length( Grab_Line ) ;
{$IFDEF FPC_OR_DELPHI}
    Parser.Free ;
{$ELSE}
    Parser.Done ;
{$ENDIF}
end ; { TSExpression_Evaluator.Evaluate }


function TSExpression_Evaluator.IsResultString : boolean ;

begin
    Result := (SSP = 1);
end ;


function TSExpression_Evaluator.Get_Result : extended ;

begin
    Get_Result := Stack[ 1 ] ;
end ;


function TSExpression_Evaluator.Get_SResult : string ;

begin
    Get_SResult := SStack[ 1 ] ;
end ;


function TSExpression_Evaluator.Get_Dump : string ;

var s : string ;

begin
    s := inttostr(SSP)+': '+SStack[ 1 ]+','+SStack[ 2 ]+','+SStack[ 3 ]+','+SStack[ 4 ]+char(13)+char(10) ;
    s := s + inttostr(trunc(stack[0]))+': '+Floattostr(stack[1])+' '+Floattostr(stack[2])+' ';
    s := s + Floattostr(stack[3])+' '+Floattostr(stack[4])+' ';
    Get_Dump := s + Floattostr(stack[5])+' '+Floattostr(stack[6]);
end ;


function TSExpression_Evaluator._Last_Error : integer ;

begin
    _Last_Error := _Error ;
end ;


function TSExpression_Evaluator.Get_Position : integer ;

begin
    Get_Position := _Position ;
end ;


function TSExpression_Evaluator.Grab_Line : string ;

begin
{$IFDEF FPC_OR_DELPHI}
    Grab_Line := Parser.Grab_Line ;
{$ELSE}
    Grab_Line := Cvt_String( Parser.Grab_Line ) ;
{$ENDIF}
end ;


procedure TSExpression_Evaluator.Put_Token( const X : string ) ;

begin
{$IFDEF FPC_OR_DELPHI}
    Parser.Put_Token( X ) ;
{$ELSE}
    Parser.Put_Token( Temp_String( X ) ) ;
{$ENDIF}
end ;


function TSExpression_Evaluator.Token : string ;

begin
{$IFDEF FPC_OR_DELPHI}
    Token := Parser.Token ;
{$ELSE}
    Token := Cvt_String( Parser.Token ) ;
{$ENDIF}
end ;


procedure TSExpression_Evaluator.Set_Parameter_Delimiter( S : string ) ;

begin
    Parameter_Delimiter := S ;
end ;


procedure TSExpression_Evaluator.Set_Decimal_Char( C : char ) ;

begin
    Decimal_Char := C ;
end ;


procedure Eval( Expression : string ; var _Result, Imaginary : Extended ;
    var E, P : Integer ) ;

var Evaluator : TSExpression_Evaluator ;

begin
    Imaginary := 0 ;
    Evaluator.Init ;
    if( Evaluator.Evaluate( Expression ) ) then
    begin
        _Result := Evaluator.Get_Result ;
        E := 0 ;
        P := 0 ;
    end else
    begin
        _Result := 0 ;
        P := Evaluator.Get_Position ;
        E := Evaluator._Last_Error ;
    end ;
    Evaluator.Done ;
end ;



begin
    Randomized := False ;
end.
