{
        Program Name : RAMDisk
        Package Name : Subroutine Library Layer 1
        Purpose      : RAM disk (memory buffer) store
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Implements a fixed-size memory store.
}

unit RAMDisk ;

interface

uses { Subroutine Library - Layer 0... }
     _UE, // TUnified_Exception
{$IFDEF CC_Regression_Test}
     _TestInt, // TTest_Interface
{$ENDIF}

     { Subroutine Library - Layer 1... }
     _Cache, // TCOM_Cache64
     _Stores, // TCOM_Store64
     UE ; // TStandard_Error_Interface


type TRAM_Disk64 = class( TCOM_Store64 )
                       public // Constructors and destructors...
                           destructor Destroy ; override ;

                       private // Instance data...
                           _Buffer : PChar ;
                           _Cache : TCOM_Cache64 ;
                           _Minimum : longint ;
                           _Size : longint ;
                           _Bytes_Read : longint ;
                           _Bytes_Written : longint ;
                           _Reads : longint ;
                           _Writes : longint ;
                           _Error_Count : longint ;
                           _Foreign_Buffer : boolean ;

                       public // API...
                           function Is_Class( Name : PChar ) : boolean ;
                               override ;

                           function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                               var UEC : TUnified_Exception ) : TStore_Address64 ;
                               override ;

                           function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                               var UEC : TUnified_Exception ) : TStore_Address64 ;
                               override ;

                           function Max_Storage : TStore_Size64 ; override ;

                           function Min_Storage : TStore_Address64 ; override ;

                           function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                               override ;

                           function Get_Read_Only : boolean ; override ;

                           function Get_Write_Only : boolean ; override ;

                           procedure Format ; override ;

                           function Get_Name : PChar ; override ;

                           function Get_Cache : TCOM_Cache64 ; override ;

                           procedure Set_Cache( Value : TCOM_Cache64 ) ;
                               override ;

                           function Contiguous_Store : boolean ; override ;

                           procedure Set_Max_Storage( Value : TStore_Address64 ;
                               var Res : TUnified_Exception ) ;
                               override ;

                           function Extended_Size : TStore_Address64 ; override ;
                           procedure Set_Read_Only( Value : boolean ) ;
                               override ;
                           procedure Set_Write_Only( Value : boolean ) ;
                               override ;

                           function Get_Bytes_Read : longint ; override ;
                           function Get_Bytes_Written : longint ; override ;
                           function Get_Reads : longint ; override ;
                           function Get_Writes : longint ; override ;
                           function Get_Error_Count : longint ; override ;
                           procedure Set_Bytes_Read( Value : longint ) ;
                               override ;
                           procedure Set_Bytes_Written( Value : longint ) ;
                               override ;
                           procedure Set_Reads( Value : longint ) ; override ;
                           procedure Set_Writes( Value : longint ) ; override ;
                           procedure Set_Error_Count( Value : longint ) ;
                               override ;
                               
                       public // API...
                           procedure Set_Buffer( P : pointer ;
                               Value : TStore_Address64 ;
                               var Res : TUnified_Exception ) ;
                           function Get_Buffer : pointer ;
                           procedure Set_Min_Storage( Value : longint ) ;
                   end ; // TRAM_Disk

{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;
{$ENDIF}

implementation

uses // Pascal....
{$IFDEF CC_Regression_Test}
     Forms,
{$ENDIF}
     Sysutils,

     // Conroy & Conroy Subroutine Library...
     CommonUt, // Edit
     Maths ; // Round_Up

const RD_Error_Creation_Failure = 1 ;
const RD_Error_Address_Out_Of_Range = 2 ;
const RD_Invalid_Pointer = 3 ;


function ERT( C : longint ) : string ;

begin
    case C of
        RD_Error_Creation_Failure : Result := 'Creation failure' ;
        RD_Error_Address_Out_Of_Range : Result := 'Address out of range' ;
        RD_Invalid_Pointer : Result := 'Invalid pointer' ;
        else Result := 'Unknown error' ;
    end ;
end ;


type tRAMDisk_Exception = class( TStandard_Error_Interface )
                              private
                                  _Code : longint ;
                                  _Text : string ;
                                  _Previous : tUnified_Exception ;

                              public
                                  function Get_Facility : longint ; override ;
                                  function Get_Facility_Version : longint ;
                                      override ;
                                  function Version : longint ; override ;
                                  function Severity : longint ; override ;
                                  function Error_Text( var Size, Typ : longint ) : PChar ;
                                      override ;
                                  function Get_Error : longint ; override ;
                                  function Get_Previous : tUnified_Exception ;
                                      override ;
                                  procedure Terminate ; override ;
                          end ;


function tRAMDisk_Exception.Get_Facility : longint ;

begin
    Result := 138 ;
end ;


function tRAMDisk_Exception.Get_Facility_Version : longint ;

begin
    Result := 10 ;
end ;


function tRAMDisk_Exception.Version : longint ;

begin
    Result := 10 ;
end ;


function tRAMDisk_Exception.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function tRAMDisk_Exception.Error_Text( var Size, Typ : longint ) : PChar ;

begin
    _Text := ERT( _Code ) ;
    Size := length( _Text ) ;
    Typ := 0 ;
    Result := PChar( _Text ) ;
end ;


function tRAMDisk_Exception.Get_Error : longint ;

begin
    Result := _Code ;
end ;


function tRAMDisk_Exception.Get_Previous : tUnified_Exception ;

begin
    Result := _Previous ;
end ;


procedure tRAMDisk_Exception.Terminate ;

begin
    Free ;
end ;



function Create_Exception( C : longint ; E : TUnified_Exception ) : tRAMDisk_Exception ;

begin
    Result := tRAMDisk_Exception.Create ;
    Result._Code := C ;
    Result._Previous := E ;
end ;



// TRAM_Disk64 methods...

// Constructors and destructors...

destructor TRAM_Disk64.Destroy ;

begin
    if( not _Foreign_Buffer ) then
    begin
        freemem( _Buffer ) ; // Only free if we allocated it
    end ;
    _Buffer := nil ;

    inherited Destroy ;
end ;


// API...

function TRAM_Disk64.Is_Class( Name : PChar ) : boolean ;

var S : string ;

begin
    S := Edit( string( Name ), 512 ) ;
    Result := ( S = 'tram_disk64' ) ;
    if( not Result ) then
    begin
        Result := inherited Is_Class( Name ) ;
    end ;
end ;


function TRAM_Disk64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Result := 0 ;
    if( _Size <= 0 ) then
    begin
        exit ;
    end ;
    Address := Round_Up( Address, _Minimum ) ;
    _Size := Round_Up( _Size, _Minimum ) ;
    if( ( Address < 0 ) or ( Address + _Size > Max_Storage ) ) then
    begin
        UEC := Create_Exception( RD_Error_Address_Out_Of_Range, nil ) ;
    end else
    begin
        UEC := nil ;
        move( _Buffer[ Address ], Data, _Size ) ;
    end ;
    Set_Last_Error( UEC ) ;
    if( UEC = nil ) then
    begin
        Result := _Size ;
        inc( _Reads ) ;
        _Bytes_Read := _Bytes_Read + _Size
    end ;
end ; // TRAM_Disk64.Read_Data


function TRAM_Disk64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Result := 0 ;
    if( _Size <= 0 ) then
    begin
        exit ;
    end ;
    Address := Round_Up( Address, _Minimum ) ;
    _Size := Round_Up( _Size, _Minimum ) ;
    if( ( Address < 0 ) or ( Address + _Size > Max_Storage ) ) then
    begin
        UEC := Create_Exception( RD_Error_Address_Out_Of_Range, nil ) ;
    end else
    begin
        UEC := nil ;
        move( Data, _Buffer[ Address ], _Size ) ;
    end ;
    Set_Last_Error( UEC ) ;
    if( UEC = nil ) then
    begin
        Result := _Size ;
        inc( _Writes ) ;
        _Bytes_Written := _Bytes_Written + _Size
    end ;
end ; // TRAM_Disk64.Write_Data


function TRAM_Disk64.Max_Storage : TStore_Size64 ;

begin
    Result := _Size ;
end ;


function TRAM_Disk64.Min_Storage : TStore_Address64 ;

begin
    Result := _Minimum ;
end ;


function TRAM_Disk64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := Max_Storage ;
end ;


function TRAM_Disk64.Get_Read_Only : boolean ;

begin
    Result := False ;
end ;


function TRAM_Disk64.Get_Write_Only : boolean ;

begin
    Result := False ;
end ;


procedure TRAM_Disk64.Format ;

begin
    // Nothing to do
end ;


function TRAM_Disk64.Get_Name : PChar ;

begin
    Result := nil ;
end ;


function TRAM_Disk64.Get_Cache : TCOM_Cache64 ;

begin
    Result := _Cache ;
end ;


procedure TRAM_Disk64.Set_Cache( Value : TCOM_Cache64 ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ; { Untested }
    end ;
    if( _Cache <> nil ) then
    begin
        _Cache.Detach ; { Untested }
    end ;
    _Cache := Value ;
end ;


function TRAM_Disk64.Contiguous_Store : boolean ;

begin
    Result := True ;
end ;


procedure TRAM_Disk64.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUnified_Exception ) ;

var P : Pointer ;

begin
    if( _Minimum = 0 ) then
    begin
        _Minimum := 16 ;
    end ;
    Value := Round_Up( Value, _Minimum ) ;
    P := _Buffer ;
    if( _Foreign_Buffer ) then
    begin
        P := allocmem( Value ) ;
        _Foreign_Buffer := False ;
    end else
    begin
        Reallocmem( P, Value ) ;
    end ;
    if( P = nil ) then
    begin
        Res := Create_Exception( RD_Error_Creation_Failure, nil ) ; { Untested }
    end else
    begin
        _Buffer := P ;
        _Size := Value ;
        Res := nil ;
    end ;
    Set_Last_Error( Res ) ;
end ; // TRAM_Disk64.Set_Max_Storage


procedure TRAM_Disk64.Set_Buffer( P : pointer ; Value : TStore_Address64 ;
    var Res : TUnified_Exception ) ; { Untested }

begin
    if( P = nil ) then
    begin
        Res := Create_Exception( RD_Invalid_Pointer, nil ) ;
    end else
    begin
        if( ( _Buffer <> nil ) and ( not _Foreign_Buffer ) ) then
        begin
            freemem( _Buffer ) ;
        end ;
        _Buffer := P ;
        _Size := Value ;
        Res := nil ;
        _Foreign_Buffer := True ;
    end ;
    Set_Last_Error( Res ) ;
end ;


function TRAM_Disk64.Get_Buffer : pointer ;

begin
    Result := _Buffer ;
end ;


procedure TRAM_Disk64.Set_Min_Storage( Value : longint ) ;

begin
    _Minimum := Value ;
end ;


function TRAM_Disk64.Extended_Size : TStore_Address64 ;

begin
    Result := Max_Storage ;
end ;


procedure TRAM_Disk64.Set_Read_Only( Value : boolean ) ;

begin
    // Do nothing
end ;


procedure TRAM_Disk64.Set_Write_Only( Value : boolean ) ;

begin
    // Do nothing
end ;


function TRAM_Disk64.Get_Bytes_Read : longint ;

begin
    Result := _Bytes_Read ;
end ;


function TRAM_Disk64.Get_Bytes_Written : longint ;

begin
    Result := _Bytes_Written ;
end ;


function TRAM_Disk64.Get_Reads : longint ;

begin
    Result := _Reads ;
end ;


function TRAM_Disk64.Get_Writes : longint ;

begin
    Result := _Writes ;
end ;


function TRAM_Disk64.Get_Error_Count : longint ;

begin
    Result := _Error_Count ;
end ;


procedure TRAM_Disk64.Set_Bytes_Read( Value : longint ) ;

begin
    _Bytes_Read := Value ;
end ;


procedure TRAM_Disk64.Set_Bytes_Written( Value : longint ) ;

begin
    _Bytes_Written := Value ;
end ;


procedure TRAM_Disk64.Set_Reads( Value : longint ) ;

begin
    _Reads := Value ;
end ;


procedure TRAM_Disk64.Set_Writes( Value : longint ) ;

begin
    _Writes := Value ;
end ;


procedure TRAM_Disk64.Set_Error_Count( Value : longint ) ;

begin
    _Error_Count := Value ;
end ;



{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    halt ;
end ;


var F : text ;

procedure Test_Unit( T : TTest_Interface ) ;

    function Log( const S : string ) : boolean ;

    begin
        append( F ) ;
        writeln( F, S ) ;
        closefile( F ) ;
        Result := T.Log_Operation( S ) ;
    end ;


var Buffer : array[ 0..511 ] of byte ;
    Loop : integer ;
    P : TStore_Address64 ;
    Test : TRAM_Disk64 ;
    UEC : TUnified_Exception ;

begin
    Test := TRAM_Disk64.Create ;
    Test.Format ;
    if( not Test.Contiguous_Store ) then
    begin
        Error ;
    end ;
    if( Test.Get_Name <> nil ) then
    begin
        Error ;
    end ;
    Test.Set_Max_Storage( 10485760, UEC ) ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Max_Storage <> 10485760 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 512 ) then
    begin
        Error ;
    end ;
    if( Test.Extended_Size <> Test.Max_Storage ) then
    begin
        Error ;
    end ;
    Test.Set_Cache( nil ) ;
    if( Test.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Extend( 104857600 ) <> Test.Max_Storage ) then
    begin
        Error ;
    end ;
    if( Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only ) then
    begin
        Error ;
    end ;
    fillchar( Buffer, 512, 0 ) ;
    if( Test.Write_Data( Buffer, 0, 512, UEC ) <> 512 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;
    fillchar( Buffer, 512, 1 ) ;
    if( Test.Read_Data( Buffer, 0, 512, UEC ) <> 512 ) then
    begin
        Error ;
    end ;
    for Loop := 0 to 511 do
    begin
        if( Buffer[ Loop ] <> 0 ) then
        begin
            Error ;
        end ;
    end ;
    if( Test.Read_Data( Buffer, 104857600, 0, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test.Write_Data( Buffer, 104857600, 0, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    Test.Free ;

    // Regression Test...
    Test := TRAM_Disk64.Create ;
    Test.Set_Max_Storage( 10485760, UEC ) ;
    while( True ) do
    begin
        case random( 5 ) of
            0: // Read Data...
               begin
                   P := random( 10485247 ) ;
                   if( Log( 'R' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   if( Test.Read_Data( Buffer, P, 512, UEC ) <> 512 ) then
                   begin
                       Error ;
                   end ;
                   if( UEC <> nil ) then
                   begin
                       Error ;
                   end ;
               end ;
            1: // Write_Data...
               begin
                   P := random( 10485247 ) ;
                   if( Log( 'W' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   fillchar( Buffer, 512, P and 255 ) ;
                   if( Test.Write_Data( Buffer, P, 512, UEC ) <> 512 ) then
                   begin
                       Test.Write_Data( Buffer, P, 512, UEC ) ;
                       Error ;
                   end ;
                   if( UEC <> nil ) then
                   begin
                       Error ;
                   end ;
                   if( Test.Read_Data( Buffer, P, 512, UEC ) <> 512 ) then
                   begin
                       Error ;
                   end ;
                   if( UEC <> nil ) then
                   begin
                       Error ;
                   end ;
                   for Loop := 0 to 511 do
                   begin
                       if( Buffer[ Loop ] <> P and 255 ) then
                       begin
                           Error ;
                       end ;
                   end ;
               end ;
            2: // Max_Storage...
               begin
                   if( Log( 'X' ) ) then
                   begin
                       exit ;
                   end ;
                   if( Test.Max_Storage <> 10485760 ) then
                   begin
                       Error ;
                   end ;
               end ;
            3: // Min_Storage...
               begin
                   if( Log( 'M' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   if( Test.Min_Storage <> 512 ) then
                   begin
                       Error ;
                   end ;
               end ;
            4: // Extend...
               begin
                   if( Log( 'E' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   if( Test.Extend( 104857600 ) <> 10485760 ) then
                   begin
                       Error ;
                   end ;
               end ;
        end ;
    end ;
    Test.Free ;
end ;
{$ENDIF}


end.
