/*
        Program Name : _Stores
        Package Name : Subroutine Library - Layer 1
        Purpose      : Storage object definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1

        Copyright (C) 1999-2009 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TStore is the base class for all storage-related classes.  It is an
        unmanaged store, meaning that it serves as a simple holder for data.  A
        store is assumed to be a contiguous storage area of a specific size.
        It has a base address of 0 (relative to store - not to memory).  A
        resize can be requested, but may not be supported.  It can be read/write
        read-only, or write-only.  This is commonly referred to as a BLOb
        (Binary Large OBject).
*/

#ifndef _STORESH
	#define _STORESH

	#include "eDefines.h"
	#include "_cache.h"
	#include "_ue.h"

typedef int TStore_Address ;
typedef int64 TStore_Address64 ;
typedef int TStore_Size ;
typedef int64 TStore_Size64 ;

const Null64 = 0 ;

class TCOM_Store : public TCommon_COM_Interface
{
    public: //API...
        virtual TStore_Address __stdcall Read_Data( void* Data, TStore_Address Address, TStore_Address _Size,
			TUnified_Exception* &UEC ) = 0;
        // Read data from specified Address in store to Data.  Read _Size bytes.  Returns success of operation in UEC.  Function returns the number of bytes actually read.

        virtual TStore_Address __stdcall Write_Data( void* Data, TStore_Address Address, TStore_Address _Size,
			TUnified_Exception* &UEC ) = 0;
        // Write data from Data to specified Address in store.  Write _Size bytes.  Returns success of operation in UEC.  Function returns the number of bytes actually written.

        virtual TStore_Size __stdcall Max_Storage() = 0;
        // Returns the maximum (current) size of the store, in bytes.  In other words, this is the highest valid address (plus 1).

        virtual TStore_Address __stdcall Min_Storage() = 0;
        //Returns the minimum atomic size of the store.  For instance, this would be the sector size for a disk.

        virtual TStore_Address __stdcall Extend( TStore_Address Amount ) = 0;
        // Requests extension of store by the amount specified.  Returns the amount actually added to store.  If Amount is 0, function returns 0 if store is extendable, and -1 if it is not.

        virtual bool __stdcall Get_Read_Only() = 0;
        // Indicates that the store can only be read from.

        virtual bool __stdcall Get_Write_Only() = 0;
        // Indicates that the store can only be written to.

        virtual void __stdcall Format() = 0;
        // Does a low-level format of the store.

        virtual char* __stdcall Get_Name() = 0;
        // Returns name of store.

		virtual TCOM_Cache* __stdcall Get_Cache() = 0;
        // Returns current cache object.

        virtual void __stdcall Set_Cache( TCOM_Cache* Value ) = 0;
        // Sets the current cache object.

        virtual bool __stdcall Contiguous_Store() = 0;
        // Returns True if the store's addresses are contiguous (non-sparse).

        virtual void __stdcall Set_Max_Storage( TStore_Address Value,
            TUnified_Exception* &Res ) = 0;
        // Sets the maximum (current) size of the store, in bytes.  In other words, this is the new highest valid address (plus 1).  If the store is not resizable, the call fails and an error code is returned in Res.  This should only be used to shrink the size of the store, not to extend it.  To extend, use the Extend method.

        virtual TStore_Address __stdcall Extended_Size() = 0;
        // Returns the maximum theoretical size that the store can be extended to.  If the store is not extendable, this returns the same value as Max_Storage.

        virtual int __stdcall Get_Bytes_Read() = 0; // Data bytes read
        virtual int __stdcall Get_Bytes_Written() = 0; // Data bytes written
        virtual int __stdcall Get_Reads() = 0; // Data read count
        virtual int __stdcall Get_Writes() = 0; // Data write count
        virtual int __stdcall Get_Error_Count() = 0; // Error count
        virtual void __stdcall Set_Bytes_Read( int Value ) = 0; // Data bytes read
        virtual void __stdcall Set_Bytes_Written( int Value ) = 0; // Data bytes written
        virtual void __stdcall Set_Reads( int Value ) = 0; // Data read count
        virtual void __stdcall Set_Writes( int Value ) = 0; // Data write count
        virtual void __stdcall Set_Error_Count( int Value ) = 0; // Error count
        virtual void __stdcall Set_Read_Only( bool Value ) = 0; // Sets the store read-only value.
        virtual void __stdcall Set_Write_Only( bool Value ) = 0;
} ; // TCOM_Store


class TCOM_Store64 : public TCommon_COM_Interface
{
    public: //API...
		virtual TStore_Address64 __stdcall Read_Data( void* Data, TStore_Address64 Address, TStore_Address64 _Size,
			TUnified_Exception* &UEC ) = 0;
        // Read data from specified Address in store to Data.  Read _Size bytes.  Returns success of operation in UEC.  Function returns the number of bytes actually read.

		virtual TStore_Address64 __stdcall Write_Data( void* Data, TStore_Address64 Address, TStore_Address64 _Size,
            TUnified_Exception* &UEC ) = 0;
        // Write data from Data to specified Address in store.  Write _Size bytes.  Returns success of operation in UEC.  Function returns the number of bytes actually written.

        virtual TStore_Size64 __stdcall Max_Storage() = 0;
        // Returns the maximum (current) size of the store, in bytes.  In other words, this is the highest valid address (plus 1).

        virtual TStore_Address64 __stdcall Min_Storage() = 0;
        // Returns the minimum atomic size of the store.  For instance, this would be the sector size for a disk.

		virtual TStore_Address64 __stdcall Extend( TStore_Address64 Amount ) = 0;
        // Requests extension of store by the amount specified.  Returns the amount actually added to store.  If Amount is 0, function returns 0 if store is extendable, and -1 if it is not.

        virtual bool __stdcall Get_Read_Only() = 0;
        // Indicates that the store can only be read from.

        virtual bool __stdcall Get_Write_Only() = 0;
        // Indicates that the store can only be written to.

        virtual void __stdcall Format() = 0;
        // Does a low-level format of the store.

        virtual char* __stdcall Get_Name() = 0;
        // Returns name of store.

        virtual TCOM_Cache64* __stdcall Get_Cache() = 0;
        // Returns current cache object.

		virtual void __stdcall Set_Cache( TCOM_Cache64* Value ) = 0;
        // Sets the current cache object.

        virtual bool __stdcall Contiguous_Store() = 0;
        // Returns True if the store's addresses are contiguous (non-sparse).

		virtual void __stdcall Set_Max_Storage( TStore_Address64 Value,
			TUnified_Exception* &Res ) = 0;
        // Sets the maximum (current) size of the store, in bytes.  In other words, this is the new highest valid address (plus 1).  If the store is not resizable, the call fails and an error code is returned in Res.  This should only be used to shrink the size of the store, not to extend it.  To extend, use the Extend method.

        virtual TStore_Address64 __stdcall Extended_Size() = 0;
        // Returns the maximum theoretical size that the store can be extended to.  If the store is not extendable, this returns the same value as Max_Storage.

      virtual int __stdcall Get_Bytes_Read() = 0; // Data bytes read
      virtual int __stdcall Get_Bytes_Written() = 0; // Data bytes written
      virtual int __stdcall Get_Reads() = 0; // Data read count
      virtual int __stdcall Get_Writes() = 0; // Data write count
      virtual int __stdcall Get_Error_Count() = 0; // Error count
	  virtual void __stdcall Set_Bytes_Read( int Value ) = 0; // Data bytes read
	  virtual void __stdcall Set_Bytes_Written( int Value ) = 0; // Data bytes written
	  virtual void __stdcall Set_Reads( int Value ) = 0; // Data read count
	  virtual void __stdcall Set_Writes( int Value ) = 0; // Data write count
	  virtual void __stdcall Set_Error_Count( int Value ) = 0; // Error count
	  virtual void __stdcall Set_Read_Only( bool Value ) = 0; // Sets the store read-only value.
	  virtual void __stdcall Set_Write_Only( bool Value ) = 0; // Sets the store write-only value.
} ; // TCOM_Store64
#endif

